#!/usr/bin/env python
"""This package provides the :class:`.MelodicClassificationPanel`, a FSLeyes
control which allows labels to be assigned to the volumes of a 4D NIFTI image.
"""

from .melodicclassificationpanel import MelodicClassificationPanel
